#include "exception.h"

#include <stdio.h>

void exception_trigger(rv32_cpu_t* cpu, uint32_t scause, uint32_t tval)
{
	// An exception can only be triggered by the CPU itself,
	// so we know we already own the mutex
	// We are in the CPU thread itself, but we need
	// the return of this function to be the beginning of
	// the cpu loop
	// To achieve that, we can just call cpu_loop (noreturn)
	// at the end of this function

	// Exceptions cannot be disabled

	// Set xCAUSE : exception cause, with interrupt bit set to null
	cpu->csr[CSR_SCAUSE] = scause;

	// Save previous interrupt enable in xSTATUS.xPIE
	if(cpu->csr[CSR_SSTATUS] & 0b10U)
		cpu->csr[CSR_SSTATUS] |= 0x80;

	// Set previous privilege mode in xSTATUS.xPP
    // TODO

	// Unset SIE (interrupt enable) bit
	cpu->csr[CSR_SSTATUS] &= ~0b10U;

	// Set privilege mode for exception handling, checking for delegation
    // TODO

	// Set xTVAL, exception-specific information related to xCAUSE
	cpu->csr[CSR_STVAL] = tval;

	// Set SEPC to instruction that caused exception
	cpu->csr[CSR_SEPC] = cpu->pc;

	// Set PC to xTVEC : exception handling code
	// xTVEC: [Base(30bits) Mode(2 bits)], address 4-byte aligned in base
	// Exceptions are not vectored (we can safely ignore mode)
	cpu->pc = cpu->csr[CSR_STVEC] & 0xFFFFFFFC;

    // Unlock cpu mutex, cpu_loop will lock it just after
	pthread_mutex_unlock(&cpu0_mutex);

    // cpu loop (attribute noreturn should erase previous stack)
	cpu_loop(cpu);
}
